// ************************************************
//    AVC.cpp
//	Sample file for SmallVR
//
//		Programa exemplo de Ambiente Virtual 
//	Colaborativo
//
//		Para compilar este fonte  necessrio 
//	montar um projeto com os seguitnes arquivos:
//	- AVC.cpp (esta arquivo)
//	- AVC_CommFuncs.cpp
//	- MemoryClient.cpp
//
//		Alm destes arquivos,  necessrio 
//	adicionar biblioteca de sockets:
//	
//		No Visual: wsock32.lib 
//		No DevCPP: libwsock32.a
//
//
//  Author: Marcio Serolli Pinho
//	Email: pinho@inf.pucrs.br		 
//	Homepages:
//		http://www.smallvr.org
//		http://grv.inf.pucrs.br
//		http://www.inf.pucrs.br/~pinho
//
//	Porto Alegre, Brazil, 2003.
// ************************************************

#ifdef WIN32
#include <windows.h>
#endif

#include <GL\gl.h>

// SmallVR header file
#include "SmVR.h"

// SmallVR header file for the 3DS file
#include <3DS\SmVR_C3DSLoader.h>

#include "AVC_ClientFunc.h"

// Root Object to represent the entire scene
SmVR_CGeometricObject *RootObject;
SmVR_CGeometricObject *UserObject, *TargetObject, *HandObject, *PointerObject;
SmVR_CGeometricObject *CubeObject, *FloorObject ;

SmVR_CGeometricObject *PartnerObject;

int MyID; // Identificador do usurio. 
// Setado de acordo com a ordem de conexo do 
// usurio no servidor de memria

double UserRotY = 0; // Rotao da cabea do usurio em Y
double PartnerRotY; // Rotao da cabea do parceiro em Y

// Definio de um conjunto de cores
GLfloat Vermelho[] = {1,0,0 };
GLfloat Azul[] = {0,0,1 };
GLfloat Verde[] = {0,1,0 };
GLfloat Amarelo[] = {1,1,0 };
GLfloat Marron[] = {0.60, 0.40, 0.12 };
GLfloat Laranja[] = { 0.98, 0.625,0.12 };
GLfloat Rosa[] = {0.98, 0.00, 0.47};
GLfloat Roxo[] = {0.60, 0.40, 0.70};

float WindowRatio;

// **********************************************************************
//  void DefineLuz(void)
//
// **********************************************************************
void DefineLight(void)
{
	// Define colors for a golden object
	GLfloat LuzAmbiente[]   = {0.2f, 0.2f, 0.2f } ;
	GLfloat LuzDifusa[]   = {0.6f, 0.6f, 0.6f, 1.0f };
	GLfloat LuzEspecular[] = {0.9f, 0.9f, 0.9f, 1.0f };
	GLfloat PosicaoLuz0[]  = {0.0f, 4.0f, 0.0f, 1.0f };
	GLfloat PosicaoLuz1[]  = {0.0f, 4.0f, 0.0f, 1.0f };
	GLfloat Especularidade[] = {1.0f, 1.0f, 1.0f, 1.0f };
	
	// ****************  Light source 0
	
	glEnable ( GL_COLOR_MATERIAL );
	
	// Enable the use of light
	glEnable(GL_LIGHTING);
	
	// Set ambient light
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, LuzAmbiente);
	// Define light 0 parameters
	glLightfv(GL_LIGHT0, GL_AMBIENT, LuzAmbiente);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LuzDifusa  );
	glLightfv(GL_LIGHT0, GL_SPECULAR, LuzEspecular  );
	glLightfv(GL_LIGHT0, GL_POSITION, PosicaoLuz0 );
	glEnable(GL_LIGHT0);
	
	// Enable the Color Tracking
	glEnable(GL_COLOR_MATERIAL);
	
	// Define the material reflectance
	glMaterialfv(GL_FRONT,GL_SPECULAR, Especularidade);
	
	// Define shininess concentration
	// As higher this value is the more concentrate
	// the shininess will be. The valid values are form 0 to 128
	glMaterialf(GL_FRONT,GL_SHININESS,51.2f);
	
	// ****************  Light source 1
	
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, LuzAmbiente);	
	
	glLightfv(GL_LIGHT1, GL_AMBIENT, LuzAmbiente);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LuzDifusa  );
	glLightfv(GL_LIGHT1, GL_SPECULAR, LuzEspecular  );
	glLightfv(GL_LIGHT1, GL_POSITION, PosicaoLuz1 );
	glEnable(GL_LIGHT1);
	
	glEnable(GL_COLOR_MATERIAL);
	
	glMaterialfv(GL_FRONT,GL_SPECULAR, Especularidade);
	
	glMateriali(GL_FRONT,GL_SHININESS,20);
	
}

#define HALF_EGDE_LENGHT 0.5
// **********************************************************************
//  void DesenhaCubo()
//
//
// **********************************************************************
void DrawCube()
{
	glBegin ( GL_QUADS );
	// Front Face
	glNormal3f(0,0,1);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	// Back Face
	glNormal3f(0,0,-1);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	// Top Face
	glNormal3f(0,1,0);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	// Bottom Face
	glNormal3f(0,-1,0);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	// Right face
	glNormal3f(1,0,0);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f( HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	// Left Face
	glNormal3f(-1,0,0);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT);
	glVertex3f(-HALF_EGDE_LENGHT,  HALF_EGDE_LENGHT, -HALF_EGDE_LENGHT);
	glEnd();
}

// **********************************************************************
// int DrawTest(void *p)
//		Desenha um cubo setando a cor de acordo com o parmetro definido
//	atravs do mtodo "SetRenderFunctionData"  
//
// **********************************************************************
int DrawTest(void *p) 
{ 
    GLfloat *Cor;
	
	Cor = (GLfloat*) p;
	
	if (Cor)
		glColor3fv(Cor);
	DrawCube(); 	
	
	return 1;
} 

// **********************************************************************
// int DrawAvatar(void *p)
//
//
// **********************************************************************
int DrawAvatar(void *p)
{
	GLfloat *Cor;
	
	Cor = (GLfloat*) p;
	
	if (Cor)
		glColor3fv(Cor);

	glPushMatrix();
		DrawCube();  // cabea 
		glPushMatrix(); // pescoo
			glTranslatef(0,-0.6,0);
			glScalef(0.5,1.0,0.5);
			DrawCube();
		glPopMatrix();
		glColor3fv(Azul);
		glPushMatrix(); // nariz
			glTranslatef(0,0.0,-0.6);
			glScalef(0.3,0.3,1.0);
			DrawCube();
		glPopMatrix();
	glPopMatrix();
	
	return 1;

}
// **********************************************************************
//  void PosicUser()
//      posiciona o observador no cenrio virtual.
//
// **********************************************************************
void PosicUser()
{
	
	SmVR_CPoint	User, Target, Zero (0,0,0);
	
	// Obtm as coordenadas do observador a partir do objeto 'UserObject'
	UserObject->GetPointInOCS(Zero, &User, RootObject);

    printf("User: %7.2f, %7.2f, %7.2f\n", User.X, User.Y, User.Z);

	// Obtm as coordenadas do alvo a partir do objeto 'TargetObject'
	TargetObject->GetPointInOCS(Zero, &Target, RootObject);
	
	// Set the clipping volume
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (90,WindowRatio,0.01,200);
	
	// Set the user Position
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();	
	gluLookAt (User.X, User.Y, User.Z,
		Target.X, Target.Y, Target.Z,
		0.0f,1.0f,0.0f);
}
// **********************************************************************
//  void reshape( int w, int h )
//		trata o redimensionamento da janela OpenGL
//
// **********************************************************************
void reshape( int w, int h )
{
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;
	
	WindowRatio = 1.0f * w / h;
	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);
	
	PosicUser();
	
}
// **********************************************************************
//  void UpdateRemoteData()
//		Faz a leitura dos dados do parceiro no AVC a partir 
//	da RemoteMemory
//
// **********************************************************************
void UpdateRemoteData()
{
	double PartnerPos[3], UserPos[3];
	SmVR_CPoint	User, Zero (0,0,0);
	
	// verifica se h um parceiro
	if (IsPartnerOnLine())
	{
		// se houver torna seu avatar visvel
		PartnerObject->SetVisibility(TRUE);
		// Obtm os dados de sua posio
		LeDadosDaMemoriaRemota(PartnerPos, PartnerRotY);
		
		printf("Posicao do Parceiro: %7.2f, %7.2f, %7.2f\n", PartnerPos[0], PartnerPos[1], PartnerPos[2]);
		// Atualiza a posio/rotaco do avatar
		PartnerObject->SetLocalTransformationMatrixToIdentity();
		PartnerObject->TranslateBy(PartnerPos[0], PartnerPos[1], PartnerPos[2]);
		PartnerObject->RotateBy(PartnerRotY,0,1,0);
	}
	
	// Envia sua prpria posio/rotaco para a RemoteMemoty 
	UserObject->GetPointInOCS(Zero, &User, RootObject);
	UserPos[0] = User.X;
	UserPos[1] = User.Y;
	UserPos[2] = User.Z;
	GravaDadosNaMemoriaRemota(UserPos, UserRotY);
	
	
}
// **********************************************************************
//  void display( void )
//		Exibe a cena
//
// **********************************************************************
void display( void )
{
	
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	UpdateRemoteData();
	
	PosicUser();
	
	glMatrixMode(GL_MODELVIEW);
	
	DefineLight();
	
	RootObject->Render();
	
	glutSwapBuffers();
}

// **********************************************************************
//  void specialFunc (int key, int x, int y) 
//		tratamento de teclas especiais
//
// **********************************************************************
void specialFunc (int key, int x, int y) 
{
	
	SmVR_CPoint	User, Zero (0,0,0);
	
	switch (key) {
	case GLUT_KEY_PAGE_UP:
		break;
		
	case GLUT_KEY_PAGE_DOWN:
		break;
		
	case GLUT_KEY_LEFT:  // Gira o corpo para esquerda
		UserRotY +=10;
		UserObject->GetPointInOCS(Zero, &User, RootObject);
		UserObject->SetLocalTransformationMatrixToIdentity();
		UserObject->TranslateBy (User);
		UserObject->RotateBy(UserRotY, 0,1,0);
		break;
		
	case GLUT_KEY_RIGHT: // Gira o corpo para direita
		UserRotY -=10;
		UserObject->GetPointInOCS(Zero, &User, RootObject);
		UserObject->SetLocalTransformationMatrixToIdentity();
		UserObject->TranslateBy (User);
		UserObject->RotateBy(UserRotY, 0,1,0);
		break;
		
	case GLUT_KEY_UP:  // Anda para frente
		// Desloca-se na direo do alvo	
		UserObject->TranslateByOnOBJCS (0,0,-1, TargetObject);
		UserObject->GetPointInOCS(Zero, &User, RootObject);
		//printf("User: %7.2f, %7.2f, %7.2f\n", User.X, User.Y, User.Z);
		break;
		
	case GLUT_KEY_DOWN: // Anda para trs 
		// Desloca-se na direo do alvo	
		UserObject->TranslateByOnOBJCS (0,0,1, TargetObject);
		UserObject->GetPointInOCS(Zero, &User, RootObject);
		//printf("User: %7.2f, %7.2f, %7.2f\n", User.X, User.Y, User.Z);
		break;
	}


}

// **********************************************************************
//  void keyboard ( unsigned char key, int x, int y )
//
//
// **********************************************************************
void keyboard ( unsigned char key, int x, int y )  
{
	switch ( key ) 
	{
	case 27:        // Teminate the program
		exit(0);    
		break;  
		// Moves the User's Body on the target CS
	case 's' : // Anda para direita
	case 'S' : 
		UserObject->TranslateByOnOBJCS (1,0,0, TargetObject);
		break;        				
		// Moves the User's Body on the target CS
	case 'a' : 
	case 'A' : // Anda para esquerda
		UserObject->TranslateByOnOBJCS (-1,0,0, TargetObject);
		break;        			
		
        // Attach the object to the pointer
	case 'p': PointerObject->AddChild(CubeObject);
		break;
	case 'l': RootObject->AddChild(CubeObject);
		break;		
		// Rotate the User's Hand
	case 'x': HandObject->RotateBy(10,1,0,0);
		break;
	case 'X': HandObject->RotateBy(-10,1,0,0);
		break;
	case 'c': HandObject->RotateBy(10,0,1,0);
		break;
	case 'C': HandObject->RotateBy(-10,0,1,0);
		break;
	}
}

// **********************************************************************
//  void SetupTheUser ()
//       Define todos os objetos que fazerm parte do "observador".
//			
// **********************************************************************
void SetupTheUser ()
{
	
	// Create the user object
	UserObject = new SmVR_CGeometricObject("User Object");
	
	// Create the target object
	TargetObject = new SmVR_CGeometricObject("Target Object");
	
	// Create a hand without Geometry
	HandObject = new SmVR_CGeometricObject("Hand Object");
	
	// Create a POINTER object
	PointerObject = new SmVR_CGeometricObject("Pointer Object", DrawTest);
	PointerObject->SetRenderFunctionData(Azul);
	
	// Add the USER to the scene
	RootObject->AddChild(UserObject);
	
	// Add the Target to the User 
	UserObject->AddChild(TargetObject);
	
	// Add the Hand to the User 
	UserObject->AddChild(HandObject);
	
	// Add the Pointer to the Hand
	HandObject->AddChild(PointerObject);
	
	// Move the USER to its initial position
	// (REMEMBER: the Target, the Hand and the Pointer are also moved !!) 
	UserObject->TranslateBy(0,1.5,5);
	
	// Move the TARGET to its initial position
	// (Remember: the translation is allways in the PARENT CS) 
	TargetObject->TranslateBy(0,0,-1);
	
	// Move the HAND to its initial position
	HandObject->TranslateBy(0.2, 0.0, -0.5);
	
	// Resize the pointer
	PointerObject->ScaleBy(0.1, 0.1,1);
	// Move the pointer to place its base on the user's hand
	PointerObject->TranslateBy(0,0,-0.5);
	
	// Create a Partner object
	
	PartnerObject = new SmVR_CGeometricObject("Partner Object", DrawAvatar);

	if (MyID == ID_CLIENT_ONE)
	{
		PartnerObject->SetRenderFunctionData(Verde);
	}
	if (MyID == ID_CLIENT_TWO)
	{
		PartnerObject->SetRenderFunctionData(Rosa);	 
		// Move the user again if he is the second one
		// move the user forward
		UserObject->TranslateToOnOBJCS(0,0,-5,TargetObject);
		// Rotate it 
		UserRotY = 180;
		UserObject->RotateBy(UserRotY, 0,1,0);		
	}
		
	RootObject->AddChild(PartnerObject);

	//Place the partner in its initial position
	PartnerObject->TranslateToOnOBJCS(0,0,0,RootObject);
	PartnerObject->ScaleBy(2,2,2);
	// Hide the partner
	PartnerObject->SetVisibility(FALSE);

}

// **********************************************************************
//  void init(void)
//		Inicializa os parmetros globais de OpenGL
//
// **********************************************************************
void InitGL(void)
{
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f); // Fundo de tela preto
	glColorMaterial ( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
	
	glShadeModel(GL_SMOOTH);						
	glColorMaterial ( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
	glEnable(GL_DEPTH_TEST);
	glEnable ( GL_CULL_FACE );	
}

// **********************************************************************
//  void main ( int argc, char** argv )
//
//
// **********************************************************************
int main ( int argc, char** argv )   
{
	
	// GLUT initialization
	glutInit            ( &argc, argv ); 
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowPosition (0,0);
	glutInitWindowSize  ( 350, 250 ); 
	glutCreateWindow    ( "Ambientes Virtuais Colaborativos" ); 
	
	// initialize OpenGL parameters
	InitGL ();
	
	// initialize the library
	RootObject = SmVR_Init(NULL);
	
	// Create a new object
	CubeObject = new SmVR_CGeometricObject("Cube Object", DrawTest);
	CubeObject->SetRenderFunctionData(Vermelho);
	
	// Add an object to the scene
	RootObject->AddChild(CubeObject);
	CubeObject->ScaleBy(0.2, 3, 0.2);
	
	// Add the floor to the scene
	// Create a new object
	FloorObject = new SmVR_CGeometricObject("Floor Object", DrawTest);
	FloorObject->SetRenderFunctionData(Marron);
	RootObject->AddChild(FloorObject);
	FloorObject ->TranslateBy(0,-1,0);
	FloorObject ->ScaleBy(10,0.1,10);
	
	// Connect to the Memory Server and get the user ID
	ConnectToTheServer(MyID);

	// Set the window caption according to the user ID
	switch(MyID)
	{
	case -1:
		glutSetWindowTitle("Cliente no conectado ao Servidor!!");
		break;
	case 0: 
		glutSetWindowTitle("Identificao no definida !!");
		break;
	case ID_CLIENT_ONE: 
		glutSetWindowTitle("Client ONE - Connected !");
		break;
	case ID_CLIENT_TWO: 
		glutSetWindowTitle("Client TWO - Connected !");
		break;
	}
	
	// Set the user, the target and the pointer
	SetupTheUser ();
	
	// GLUT stuff
	glutDisplayFunc ( display );  
	glutReshapeFunc ( reshape );
	glutKeyboardFunc ( keyboard );
	glutSpecialFunc (specialFunc);
	glutIdleFunc ( display );
	
	glutMainLoop ( );          
	return 0;	
}


