/*************************************************
 **
 ** Utilities.h
 ** Library with color and other stuffs.
 **
 ** Author: Rafael Saraiva Garcia.
 ** E-Mail: rafael_sgarcia@yahoo.com
 ** Version: 1.0
 ** Date: July 31th, 2002.
 **
 *************************************************/
#include "Utilities.h"



/*************************************************
 ** Find Point.
 *************************************************/
void findPoint (Point3D *inPoint, Point3D *outPoint) {

	GLfloat newPoint[4];
	GLfloat glMatrix[4][4];

	glGetFloatv (GL_MODELVIEW_MATRIX, &glMatrix[0][0]);

	for (int counter = 0; counter < 4; counter++) {
		newPoint[counter] = glMatrix[0][counter] * inPoint->x +
			glMatrix[1][counter] * inPoint->y +
			glMatrix[2][counter] * inPoint->z +
			glMatrix[3][counter];
	}

	outPoint->x = newPoint[0];
	outPoint->y = newPoint[1];
	outPoint->z = newPoint[2];

}



/*************************************************
 ** Change Resolution.
 *************************************************/
void changeResolution (int mode) {

	DEVMODE devMode;
	int ret;

	devMode.dmSize = sizeof (DEVMODE);
	EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &devMode);

	switch (mode) {
		case RES_640x480:
			devMode.dmPelsWidth = 640;
			devMode.dmPelsHeight = 480;
			break;

		case RES_800x600:
			devMode.dmPelsWidth = 800;
			devMode.dmPelsHeight = 600;
			break;

		case RES_1024x768:
			devMode.dmPelsWidth = 1024;
			devMode.dmPelsHeight = 768;
			break;

		case RES_1400x1050:
			devMode.dmPelsWidth = 1400;
			devMode.dmPelsHeight = 1050;
			break;

		default:
			printf ("Invalid vide mode!!!\n\n");
			return;
	}

	ret = ChangeDisplaySettings (&devMode, CDS_FULLSCREEN | CDS_TEST);

	switch (ret) {
		case DISP_CHANGE_SUCCESSFUL:
			printf ("The settings change was successfull.\n");
			ret = ChangeDisplaySettings (&devMode, CDS_FULLSCREEN);
			switch (ret) {
				case DISP_CHANGE_SUCCESSFUL:
					printf ("Resolution successfully changed!\n");
					break;

				case DISP_CHANGE_RESTART:
					printf ("A rebbot is necessary before the changes will take effect.\n");
					break;

				default:
					printf ("Unable to change resolution!\n");
			}
			break;

		case DISP_CHANGE_RESTART:
			printf ("The computer must be restarted in order for the graphics mode to work.\n");
			break;

		
		case DISP_CHANGE_BADFLAGS:
			printf ("An invalid set of flags was passed in.\n");
			break;

		case DISP_CHANGE_BADPARAM:
			printf ("An invalid parameter was passed in.\n");
			break;

		case DISP_CHANGE_FAILED:
			printf ("The display driver failed the specified graphics mode.\n");
			break;

		case DISP_CHANGE_BADMODE:
			printf ("The graphics mode is not supported.\n");
			break;

		case DISP_CHANGE_NOTUPDATED:
			printf ("Unable to write settings to the registry.\n");
			break;
	}

}



/*************************************************
 ** Set Window Titles.
 *************************************************/
void setWindowTitles (char *title) {

	char auxiliar[200];
	strcpy (auxiliar, title);

	SetConsoleTitle (auxiliar);
	SetWindowText (GetActiveWindow(), title);
}



/*************************************************
 ** Remove Border.
 *************************************************/
void removeBorder (void) {
	static int IS_FULL_NOW = FALSE;
	double error;

	LONG style;
	static LONG oldStyle;

	HWND hwnd;

	changeResolution (RES_640x480);
	hwnd = GetActiveWindow();
	error = GetLastError();

	if (error)
		printf ("0 - Erro: %f.\n", error);

	if (IS_FULL_NOW) {
		style = oldStyle;
		IS_FULL_NOW = TRUE;
	}
	else {
		oldStyle = GetWindowLong (hwnd, GWL_STYLE);
		style = WS_POPUP | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
		IS_FULL_NOW = FALSE;
	}

	SetWindowLong (hwnd, GWL_STYLE, style);
}



/*************************************************
 ** Set Color.
 *************************************************/
void setColor (GLfloat red, GLfloat green, GLfloat blue) {
	glColor3f (red, green, blue);
}



/*************************************************
 ** Set Background Color.
 *************************************************/
void setBackgroundColor (GLfloat red, GLfloat green, GLfloat blue) {
	glClearColor (red, green, blue, 1.0);
}



/*************************************************
 ** Reshape Function.
 *************************************************/
void reshapeFunction (GLsizei width, GLsizei height, GLfloat range) {
	// Prevent division by zero.
	if (height == 0)
		height = 1;

	// Set viewport to windows dimension.
	glViewport (0, 0, width, height);

	// Reset projection coordinate system.
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();

	// Establish clipping volume.
	if (width <= height)
		glOrtho (-range, range, -range * height / width, range * height / width, -range, range);
	else
		glOrtho (-range * width / height, range * width / height, -range, range, -range, range);

	// Reset modelview coordinate system.
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity();
}



/*************************************************
 ** Keyboard Function.
 *************************************************/
void keyboardFunction (unsigned char key, int x, int y, GLfloat *xRotate, GLfloat *yRotate, GLfloat *zRotate) {
	switch (key) {
		// Numpad 01.
		case 49: 
			*xRotate += 5.0; 
			*yRotate -= 5.0; 
			break;

		// Numpad 02.
		case 50: 
			*xRotate += 5.0;				   
			break;

		// Numpad 03.
		case 51: 
			*xRotate += 5.0; 
			*yRotate += 5.0; 
			break;

		// Numpad 04.
		case 52: 
			*yRotate -= 5.0;				   
			break;

		// Numpad 05.
		case 53: 
			*xRotate = 0.0; 
			*yRotate = 0.0; 
			*zRotate = 0.0; 
			break;

		// Numpad 06.
		case 54: 
			*yRotate +=5.0; 
			break;

		// Numpad 07.
		case 55: 
			*xRotate -= 5.0; 
			*yRotate -= 5.0; 
			break;

		// Numpad 08.
		case 56: 
			*xRotate -= 5.0; 
			break;

		// Numpad 09.
		case 57: 
			*xRotate -= 5.0; 
			*yRotate += 5.0;
			break;

		default:
			break;
	}

	glutPostRedisplay();
}



/*************************************************
 ** Special Keyboard Function.
 *************************************************/
void specialKeyboardFunction (int key, int x, int y, GLfloat *xRotate, GLfloat *yRotate, GLfloat *zRotate) {
	switch (key) {
		// Left Key.
		case GLUT_KEY_LEFT:  
			*yRotate -= 5.0; 
			break;

		// Right Key.
		case GLUT_KEY_RIGHT: 
			*yRotate += 5.0; 
			break;

		// Up Key.
		case GLUT_KEY_UP:    
			*xRotate -= 5.0; 
			break;

		// Down Key.
		case GLUT_KEY_DOWN:  
			*xRotate += 5.0; 
			break;

		default:
			break;
	}

	glutPostRedisplay();
}
